# Copyright 2009 Canonical Ltd.
#
# This file is part of wsgi-oops.
#
# wsgi-oops is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# wsgi-oops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wsgi-oops.  If not, see <http://www.gnu.org/licenses/>.


"""
The OOPS serializer is responsible for allocating OOPS IDs and
dumping an OOPS into persistent media (disk, database, whatever).
It must store versioning information, contained on each of the
OOPS and OOPS entries, so that the precise format may be changed
over time.

OOPS IDs are composed as follows:

<DAYS SINCE EPOCH><KEY><SEQUENCE>

Where:

   - <DAYS SINCE EPOCH> is an integer computed from the number of days
     elapsed since 2006-01-01.

   - <KEY> is a letters-only parameter passed in to the constructor.

   - <SEQUENCE> is an integer incremented each time an OOPS ID is
     generated.  The is reset to 0 when <DAYS SINCE EPOCH> changes,
     and must be unique for each given <DAYS SINCE EPOCH> + <KEY> pair.

This means that:

854FOO67

Was the 67th OOPS ID generated for the FOO key on 2008-05-04.
"""

from datetime import datetime, timedelta

import os
import pytz
import re
import subprocess
import threading
import weakref
import zipfile

from pprint import pformat
from canonical.versioninfo import version_info

UTC = pytz.utc
EPOCH = datetime(2006, 01, 01)


class OOPSSerializer(object):
    """
    Oops Serialiser object, a base class that should be extended
    with methods for storing the data in specific ways.
    """

    _sequences = {}
    _sequences_day = None
    _mutex = threading.Lock()
    _serializers = {}

    def __init__(self, key):
        if not re.match("^[a-zA-Z]+$", key):
            raise ValueError("Invalid OOPS key: %s" % (key,))

        # We don't want unicode OOPS IDs.
        if isinstance(key, unicode):
            key = str(key)

        self._key = key

        self._mutex.acquire()
        try:
            # We want unique OOPS ids across all serializers, so we keep
            # track of serializers created, and query them for the initial
            # sequence.  The initial sequence used is the maximum value
            # among them.
            if key not in self._serializers:
                self._serializers[key] = weakref.WeakKeyDictionary()
            self._serializers[key][self] = True

            # Force current sequences to be recomputed.
            if key in self._sequences:
                del self._sequences[key]
        finally:
            self._mutex.release()

    @classmethod
    def reset_sequences(cls):
        """Reset global sequence state.  Useful for tests."""
        OOPSSerializer._mutex.acquire()
        try:
            OOPSSerializer._sequences.clear()
            OOPSSerializer._sequences_day = None
        finally:
            OOPSSerializer._mutex.release()

    @classmethod
    def reset_serializers(cls):
        """Reset knowledge about previous serializers.  Useful for tests.
        """
        OOPSSerializer._mutex.acquire()
        try:
            OOPSSerializer._serializers.clear()
        finally:
            OOPSSerializer._mutex.release()

    def get_initial_sequence(self):
        """Return the sequence to be *used* for the given key in the next OOPS.

        When the serializer instance is recreated due to a system reset
        or some other event, the sequencing should continue where it last
        stopped.

        This method should be overwritten by subclasses.

        WARNING: Subclasses shouldn't do anything in the implementation of
        this method that would reentrantly generate an OOPS ID,
        otherwise they'll end up dead locked.
        """
        return 0

    def get_next_sequence(self):
        """Return pair (<days since epoch>, <sequence value>) to compute an ID.

        Sequence values should be unique for each key within a given day,
        even if the system is restarted, and they should be reset once the
        day changes.
        """
        # this day format (adding one) matches LP's epoch calcs.
        day = (datetime.utcnow() - EPOCH).days + 1
        self._mutex.acquire()
        try:
            if day != self._sequences_day:
                # We've got a new day.  Reset the sequencing entirely.
                # Note that we'll still ask for the initial sequence from
                # each involved instance, which means that this works even
                # if faced with clock skews.
                self._sequences.clear()
                OOPSSerializer._sequences_day = day
            if self._key not in self._sequences:
                # Key wasn't seen yet.  Use the maximum sequence from all
                # known serializers.
                sequence = max(serializer.get_initial_sequence() for
                               serializer in self._serializers[self._key])
            else:
                # That's the most common path by far, so even though this
                # all sounds complex, the overhead isn't significant.
                sequence = self._sequences[self._key]
            self._sequences[self._key] = sequence + 1
        finally:
            self._mutex.release()
        return day, sequence

    def get_next_oops_id(self):
        """Return the next unique OOPS ID."""
        day, sequence = self.get_next_sequence()
        return "%d%s%d" % (day, self._key, sequence)

    def dump(self, oops):
        """Serialize the given OOPS."""
        raise NotImplementedError("Please implement %s.dump()" %
                                  self.__class__.__name__)


class OOPSFileSerializer(OOPSSerializer):
    """
    OOPS serializer that dumps OOPSes to disk.
    """
    version = 2

    def __init__(self, key, dump_dir, fh=None):
        super(OOPSFileSerializer, self).__init__(key)
        self._dump_dir = dump_dir
        self._fh = fh
        if not os.path.isdir(dump_dir):
            os.makedirs(dump_dir)

    def get_initial_sequence(self):
        """Compute initial sequence, considering any previously dumped files.
        """
        sequence = 0
        date_path = self._get_date_path(self._sequences_day)
        if os.path.isdir(date_path):
            for filename in os.listdir(date_path):
                oops_id = filename[6:].split(".", 1)[0]
                oops_id_sequence = self._split_oops_id(oops_id)[2]
                if oops_id_sequence >= sequence:
                    sequence = oops_id_sequence + 1
        return sequence

    def _get_oops_filename(self, oops_id):
        """Return an OOPS filename like DUMPDIR/YYYY-MM-DD/SSSSS-OOPSID.oops
        """
        day = self._split_oops_id(oops_id)[0]
        now = datetime.utcnow()
        midnight = now.replace(hour=0, minute=0, second=0)
        return os.path.join(self._get_date_path(day),
                            "%05d-%s.oops" % ((now-midnight).seconds, oops_id))

    def _start_dump(self, oops):
        """Start a dump by making sure the directory exists"""
        filename = self._get_oops_filename(oops.assign_id())
        oops_dir = os.path.dirname(filename)
        try:
            os.makedirs(oops_dir)
        except OSError:
            # We must test *after* blowing up because otherwise we
            # create a race condition (A test, B create, A create, BOOM).
            if not os.path.isdir(oops_dir):
                raise
        if os.path.isfile(filename):
            os.remove(filename)

        return filename

    def dump(self, oops):
        """Dump OOPS to disk."""
        dump_filename = self._start_dump(oops)

        # We need to pass the filename as parameter to avoid
        # race conditions between threads. Same for an oops.
        self._write_file(dump_filename, oops)
        self._finish_dump(dump_filename)

        return dump_filename

    def _write_file(self, filename, oops):
        """Basic writing data to the file system under a directory path."""
        fh = self._open_dump_file(filename)

        section = "version"
        data = self._generate_version(oops)
        fh.write("==== %s ====\n" % section)
        fh.write(str(data).strip()+"\n\n")
        for entry in oops.get_all():
            section = entry.entry_name
            data = entry.get_dump_data()
            fh.write("==== %s ====\n" % section)
            fh.write(str(data).strip()+"\n\n")

        # if using instance's fh do not close it (may be stderr)
        if self._fh is None:
            fh.close()

    def _finish_dump(self, filename=None):
        """Clear up any variables used for dumping the oops log."""
        pass

    def _open_dump_file(self, filename):
        """Returns a file handler to dump to."""
        if self._fh:
            fh = self._fh
        elif os.path.isdir(filename):
            filepath = os.path.join(filename, section)
            fh = open(filepath, "w")
        else:
            filepath = filename
            fh = open(filepath, "a")

        return fh

    def _generate_version(self, oops):
        """Generate a textual file with versioning information."""
        lines = []
        lines.append("oops %d\n" % oops.version)
        lines.append("serializer %d\n" % self.version)
        for entry in oops.get_all():
            lines.append("%s %d\n" % (entry.entry_name, entry.entry_version))
        lines.sort()
        return { 'version': "".join(lines) }

    def _get_date_path(self, day):
        """Return a path like DUMPDIR/YYYY-MM-DD for the given epoch-based day.
        """
        # subtract one from the earlier calculated lp style day to get the dir
        # to match the oops date.
        oops_datetime = EPOCH + timedelta(days=day-1)
        return os.path.join(self._dump_dir, oops_datetime.strftime("%Y-%m-%d"))

    def _split_oops_id(self, oops_id):
        """Split OOPS ID into its (day, key, sequence) components."""
        match = re.match("^(\d+)([a-zA-Z]+)(\d+)$", oops_id)
        if not match:
            raise ValueError("Invalid OOPS ID: %s" % (oops_id,))
        day, key, sequence = match.groups()
        return int(day), key, int(sequence)


class OOPSZipSerializer(OOPSFileSerializer):
    """Outputs the oops files into a zip file."""
    def __init__(self, key, dump_dir):
        self._zip = None
        super(OOPSZipSerializer, self).__init__(key, dump_dir)

    def _write_file(self, filename, oops):
        """Write a single oops data section to the zipfile"""

        zip_handler = zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED)
        prefix = "%s/" % oops.id

        section = "version"
        data = self._generate_version(oops)
        zip_handler.writestr(os.path.join(prefix, section), str(data))

        for entry in oops.get_all():
            section = entry.entry_name
            data = entry.get_dump_data()
            zip_handler.writestr(os.path.join(prefix, section), str(data))

        zip_handler.close()


class OOPSRFC822Serializer(OOPSFileSerializer):
    """Outputs the oops files into an rfc822."""

    def _normalise(self, s):
        """Normalise the whitespace in a string to spaces"""
        if s is None:
            return None
        return ' '.join(s.split())

    def _write_header(self, fh, all_headers):
        """Write rfc822 header to a fh from an array of name/value pairs."""
        for headers in all_headers:
            for header in headers:
                value = self._normalise(str(header[1]))
                fh.write('%s: %s\n' % (header[0], value))
        fh.write('\n')

    def _write_file(self, filename, oops):
        """Write a single oops data section to the rfc822 file"""
        data = {}
        data["version"] = self._generate_version(oops)
        for entry in oops.get_all():
            data[entry.entry_name] = entry.get_dump_data()

        fh = self._open_dump_file(filename)

        bits = self._get_bits( data )
        headers = ([
            # XXX: assign_id? what if the oops already has it?
            # and I think the latter *is* the case
            ( 'Oops-Id', oops.assign_id() ),
            ( 'Exception-Type', bits['exception'][0] ),
            ( 'Exception-Value', bits['exception'][1] ),
            ( 'Date', datetime.now(UTC).isoformat() ),
            ( 'Branch', version_info['branch_nick'] ),
            ( 'Revision', version_info['revno'] ),
            ( 'Last-Update', version_info['date'] ),
            ( 'URL', bits['url'] ),
            ( 'HTTP-Status', bits['status'] ),
            ( 'Duration', bits.get('dur', 'unknown')),
        ], self._get_session_info(bits.get('session')))
        self._write_header(fh, headers)

        if data.has_key('metadata'):
            for name, value in data.pop('metadata').iteritems():
                fh.write("%s=%s\n" % (name, str(value)))
            fh.write("\n")
        if data.has_key('storm'):
            storm = data.get('storm')
            last = 0
            for s in storm.get('statements', []):
                if s[3] is None or s[2] is None:
                    size = 0
                else:
                    size = "%0.4f" % (s[3] - s[2])
                    size = int(size.split('.')[1])
                next = last + size
                fh.write("%04d-%04d %s\n" % (last, next, s[0]))
                last = next
            if last > 0:
                data.pop('storm')
                fh.write("\n")

        if bits.has_key('traceback') and bits['traceback']:
            fh.write("Traceback (most recent call last):\n")
            for frame in bits.get('traceback', []):
                fh.write("  Module %s, line %d, in %s\n    %s\n" % (
                frame['file'], frame['lineno'],
                frame['method'], frame['line']))
            fh.write("%s: '%s'\n" % (
                bits['exception'][0], bits['exception'][1]))
            fh.write("\n")

        fh.write("== EXTRA DATA ==\n")
        fh.write(pformat(data)+"\n")

        # if using instance's fh do not close it (may be stderr)
        if self._fh is None:
            fh.close()

    def _get_bits(self, data):
        """
        Parse the mirad of data passed in to take out interesting parts
        for the rfc headers.
        """
        result = {}
        result['exception'] = ['Unknown', 'None']
        result['traceback'] = []
        log = data.get('log', {})
        if log and log.has_key('records'):
            for record in log['records']:
                if record.has_key('args'):
                    args = record.pop('args')
                    if record['levelname'] == 'INFO':
                        msg = record.get('message', record.get('msg', None))
                        if msg == 'Session':
                            result['session'] = args
                if record.has_key('oops_traceback'):
                    trace = record.pop('oops_traceback')

                    if trace['exception']:
                        # just a quick fix. I'll fix this in a later patch,
                        # as the problem is deeper (the exception to str
                        # and back to exception is wrong!)
                        etype_value = trace['exception'].split(":", 1)
                        if len(etype_value) == 2:
                            etype, value = etype_value
                        else:
                            etype, value = etype_value[0], 'unknown'
                        result['exception'] = [etype, value]

                    for frame in trace.get('frames', []):
                        # Check to see if we've hit oops it's self
                        if 'oops/wsgi.py' in frame[0]:
                            break
                        result['traceback'].append( {
                            'file': frame[0],
                            'lineno': frame[1],
                            'method': frame[2],
                            'line': frame[3],
                            'locals': frame[4],
                        } )

        meta = data.get('metadata', {})
        for key in meta.keys():
            if type(meta[key]) not in [str, unicode, int, float]:
                meta.pop(key)

        # Be careful, sometimes we write oops even though we are not in a wsgi
        # stack.
        result['url'] = meta.get('PATH_INFO', 'No PATH_INFO available')
        if meta.has_key('request-start'):
            result['dur'] = float(meta['request-end'])-float(meta['request-start'])
        result['status'] = meta.get('http-code', 'No http-code available')
        return result

    def _get_session_info(self, session):
        """Get session information from the session info log"""
        if session:
            return [
                ( 'User-id', session.get('user') ),
                ( 'User-name', session.get('name') ),
                ( 'Session-id', session.get('session') ),
            ]
        else:
            return list()

class OOPSBz2Serializer(OOPSRFC822Serializer):
    """A serializer that will generate compressed oops files."""

    def _finish_dump(self, filename):
        # have the parent finishing the plain oops dump
        super(OOPSBz2Serializer, self)._finish_dump(filename)

        # compress it only if filename it's a existent file
        if os.path.exists(filename):
            subprocess.check_call(('bzip2', filename))
