# Copyright 2009 Canonical Ltd.
#
# This file is part of wsgi-oops.
#
# wsgi-oops is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# wsgi-oops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wsgi-oops.  If not, see <http://www.gnu.org/licenses/>.

"""Test canonical.oops.serializer"""

__metaclass__ = type

import bz2
import logging
import os
import rfc822
import sys
import unittest
import zipfile

from cStringIO import StringIO

from canonical.oops.oops import (
    OOPS, OOPSMetaData, OOPSLog, OOPSLoggingHandler )
from canonical.oops.serializer import (
    OOPSRFC822Serializer, OOPSZipSerializer, OOPSBz2Serializer )
from storm.tracer import install_tracer
from canonical.oops.stormtracer import OOPSStorm, OOPSStormTracer

environ = {
    'request-start' : 0,
    'request-end' : 1,
    'PATH_INFO' : 'foo/bar',
    'http-code' : '500',
    'content-type' : 'text/html',
}
session_info = {
    'user' : '101',
    'name' : 'test-user',
    'session' : 549855,
}

class OopsSerializerTests(unittest.TestCase):
    """Test oops serialisers"""
    serializer_class = OOPSRFC822Serializer

    def _get_result_as_msg(self):
        result = StringIO()
        # Tell the serializer to put data into my own fh
        s = self.serializer_class("key", "/tmp", fh=result)
        # This is where the serialization is done
        s.dump(self.oops)
        # Seek to the start of the serialized string
        result.seek(0)
        # Load the result up as an rfc822 message and close fh
        return rfc822.Message(result), result

    def setUp(self):
        """Init this test"""
        self.oops = OOPS( self.get_id )
        self.metadata = OOPSMetaData(environ)
        install_tracer(OOPSStormTracer())
        root_logger = logging.getLogger()
        root_logger.addHandler(OOPSLoggingHandler())
        root_logger.setLevel(0)

        self.oops.add(self.metadata)
        self.oops.add(OOPSLog())
        self.oops.add(OOPSStorm())

        try:
            raise ValueError, "Die!"
        except ValueError:
            logging.info("Session", session_info)
            logging.error("Test Error", exc_info=sys.exc_info())

        self.msg, result = self._get_result_as_msg()

        self.sections = [ '' ]
        section = 0
        line = self.msg.fp.readline()
        while line:
            line = line.strip()
            if line == '':
                section += 1
                self.sections.append('')
            else:
                self.sections[section] += line + "\n"
            line = self.msg.fp.readline()
        result.close()

    def get_id(self):
        """Return fixed oops id"""
        return "4000testoopsid1"

    def test_error_headers(self):
        """Test for rfc822 headers we know the value of"""
        msg = self.msg
        self.assertEqual(msg.getheader('oops-id'), self.get_id())
        self.assertEqual(msg.getheader('Exception-Type'), 'ValueError')
        self.assertEqual(msg.getheader('Exception-Value'), 'Die!')
        self.assertEqual(msg.getheader('url'), 'foo/bar' )
        self.assertEqual(msg.getheader('http-status'), '500')
        self.assertEqual(msg.getheader('duration'), '1.0')

    def test_session_headers(self):
        """Test headers which record the session variables"""
        msg = self.msg
        self.assertEqual(msg.getheader('user-id'), '101')
        self.assertEqual(msg.getheader('user-name'), 'test-user')
        self.assertEqual(msg.getheader('session-id'), '549855')

    def test_variable_headers(self):
        """Make sure all variable headers at least exist"""
        self.assert_(self.msg.getheader('date'))
        self.assert_(self.msg.getheader('branch'))
        self.assert_(self.msg.getheader('revision'))
        self.assert_(self.msg.getheader('last-update'))

    def test_environ_content(self):
        """Make sure all environment headers are saved"""
        env = self.sections[0]
        for line in env.split():
            name, value = line.split("=")
            if environ.has_key(name.upper()):
                name = name.upper()
            self.assertEqual(value, str(environ[name]))

    def test_storm_content(self):
        """Make sure sql content is stored right"""
        storm = self.sections[1]
        # No way to perform test storm execution yet
        #self.assertEqual("No it doesn't", "Yes it does")

    def test_traceback_content(self):
        """Make sure a traceback is recorded"""
        trace = self.sections[1] # Incriment when storm is enabled
        self.assert_("Die!" in trace)
        self.assert_("setUp\nraise ValueError" in trace)


class OopsBz2SerializerTests(OopsSerializerTests):
    """Run the same set of tests with the Bz2 serializer."""
    serializer_class = OOPSBz2Serializer

    def _get_result_as_msg(self):
        # Tell the serializer to dump data to a BZ file
        s = self.serializer_class("key", "/tmp")

        # This is where the serialization is done
        filename = s.dump(self.oops)
        assert filename is not None

        # store filename to remove it at tearDown time
        self.compressed_filename = filename + '.bz2'

        self.assertTrue(os.path.exists(self.compressed_filename))
        self.assertFalse(os.path.exists(filename))

        fp = bz2.BZ2File(self.compressed_filename)
        return rfc822.Message(fp), fp

    def tearDown(self):
        do_remove = hasattr(self, 'compressed_filename') and \
                    os.path.exists(self.compressed_filename)
        if do_remove:
            os.remove(self.compressed_filename)

        super(OopsSerializerTests, self).tearDown()


class OopsBz2SerializerToStderrTests(OopsSerializerTests):
    """Run the same set of tests with the Bz2 serializer using stderr as fh."""
    serializer_class = OOPSBz2Serializer

    def _get_result_as_msg(self):
        # fake stderr
        self.faked_stderr = open('/tmp/sys.stderr', 'w+', buffering=0)
        os.dup2(self.faked_stderr.fileno(), sys.stderr.fileno())

        # Tell the serializer to dump data to stderr
        s = self.serializer_class("key", "/tmp", fh=sys.stderr)

        # This is where the serialization is done
        filename = s.dump(self.oops)

        self.faked_stderr.flush()
        self.faked_stderr.seek(0)
        return rfc822.Message(self.faked_stderr), self.faked_stderr

    def tearDown(self):
        do_remove = hasattr(self, 'faked_stderr') and \
                    os.path.exists(self.faked_stderr.name)
        if do_remove:
            os.remove(self.faked_stderr.name)

        super(OopsSerializerTests, self).tearDown()


if __name__ == '__main__':
    unittest.main()
