# Copyright 2009 Canonical Ltd.
#
# This file is part of wsgi-oops.
#
# wsgi-oops is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# wsgi-oops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wsgi-oops.  If not, see <http://www.gnu.org/licenses/>.

""" Test cases and utilities for WSGI Oops """
import os
import tempfile
import unittest
import shutil
import sys

from canonical.oops import wsgi
from paste.fixture import TestApp


class OopsWareInterceptor(object):
    """ A WSGI middleware to intercet the
    canonical.oops.wsgi.ResponseTrapper instance recieved as start_response
    from canonical.oops.wsgi.OopsWare
    """

    def __init__(self, app):
        self.app = app
        self.oops = None

    def __call__(self, environ, start_response):
        if isinstance(start_response, wsgi.ResponseTrapper):
            self.oops = start_response.oops
        return self.app(environ, start_response)


class WSGIOopsTestCase(unittest.TestCase):
    """ base test case for WSGI oops related tests

    Provides some utilities to ease the writting of tests.
    i.e:
      - create a temp dir to store the oops for each test method
      - oops property to get the Oops object from the OopsWare
        (via OopsWareInterceptor)
      - provide a simple wsgi app for testing, and allow subclasses to create
        their own overriding create_test_app method
    """

    def setUp(self):
        super(WSGIOopsTestCase, self).setUp()
        self.test_root_dir = None
        self.tmp_dir = self._make_tmp_dir()
        self.oops_ware_interceptor = OopsWareInterceptor(self.create_test_app())
        self.app = wsgi.OopsWare(self.oops_ware_interceptor, self.tmp_dir)
        self.testapp = TestApp(self.app)

    def _make_tmp_dir(self):
        """ creates and returns safe temp dir for the current test method """
        if getattr(self, 'test_root_dir', None) is None:
            self.test_root_dir = tempfile.mkdtemp('', 'wsgi_oops')
        return os.path.join(self.test_root_dir, self._testMethodName)

    def tearDown(self):
        super(WSGIOopsTestCase, self).tearDown()
        self.rm_tmp_dir()

    def rm_tmp_dir(self):
        """ clean up the temp dir """
        root_dir = getattr(self, 'test_root_dir', None)
        if root_dir and os.path.exists(root_dir):
            shutil.rmtree(root_dir)

    def create_test_app(self):
        """ return a simple wsgi app """
        return SimpleApplication()

    @property
    def oops(self):
        """ the Oops instance of the current request """
        return self.oops_ware_interceptor.oops


class SimpleApplication(object):
    """ Simple WSGI application that allows raise forced errors """

    def __call__(self, environ, start_response):
        query = environ.get('QUERY_STRING', '')
        if query.find('error=') >= 0:
            try:
                raise ValueError(query.split('error=')[1])
            except ValueError, ve:
                start_response("500 Unexpected server error",
                            [('Content-Type', 'text/html'),], sys.exc_info())
                return ["An error was requested"]
        else:
            start_response("200 OK", [('Content-Type', 'text/html'), ])
            return ["Hooary, all OK!"]


