# Copyright 2009 Canonical Ltd.
#
# This file is part of wsgi-oops.
#
# wsgi-oops is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# wsgi-oops is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wsgi-oops.  If not, see <http://www.gnu.org/licenses/>.

""" WSGI Middleware Wrapper for oops

Use like so::

 app = OopsWare(app, oops_dir='/tmp/oops/')

oops_dir is a directory that will be created to dump oops in to.

"""

__metaclass__ = type
__all__ = ["OopsWare"]

import sys
import time
import logging

from storm.tracer import install_tracer, remove_tracer_type

from canonical.oops.serializer  import OOPSRFC822Serializer
from canonical.oops.oops        import (
    OOPS, OOPSLog, OOPSMetaData, OOPSLoggingHandler )
from canonical.oops.stormtracer import OOPSStorm, OOPSStormTracer


class OopsWare(object):
    """
    Intercept __call__ and act accordingly, returning the
    required page unless there was an error.
    """
    def __init__(self, app, oops_dir, debug=False):
        """
        Initialise the middleware for oops.
        """
        self.app = app
        # Add the oops to the storm tracing
        remove_tracer_type(OOPSStormTracer)
        install_tracer(OOPSStormTracer())
        # Add the oops to the standard logging
        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            if isinstance(handler, OOPSLoggingHandler):
                break
        else:
            root_logger.addHandler(OOPSLoggingHandler())
        fh = debug and sys.stderr or None
        self.serial = OOPSRFC822Serializer("appserver", oops_dir, fh=fh)

    def __call__(self, environ, start_response):
        """
        Take the start of the request, record the time/date and store
        The object for later.
        """
        response_trapper = ResponseTrapper(self.serial, environ, start_response)
        return self.app(environ, response_trapper)


class ResponseTrapper(object):
    """Looks for responses in the wsgi middleware stack with which to act."""

    def __init__(self, serial, environ, start_response):
        self.serial = serial
        self.environ = environ
        # Request start, record the time and create oops tracking object
        self.oops = OOPS(self.serial.get_next_oops_id)
        self.oops.add(OOPSStorm(thread_local=True))
        self.oops.add(OOPSLog(thread_local=True))
        self.metadata = OOPSMetaData(environ)
        self.metadata["request-start"] = time.time()
        self.oops.add(self.metadata)
        self.response = start_response
        environ['OOPSID'] = self.oops.assign_id()

    def __call__(self, status, response_headers, exc_info=None):
        """
        Grab the response code and count by group or specific as needed
        """
        status_code = status[0:3]
        if status_code == '500':
            self.oops.keep()

        # This prevents random 404ed requests from logging
        if status_code == '404':
            if self.environ.get('HTTP_REFERER', None):
                self.oops.keep()

        if not exc_info or not exc_info[0]:
            exc_info = sys.exc_info()

        # Force the dumping of an oops if set.
        self.forced = self.environ.has_key('dump-oops')
        if self.oops:
            self.metadata["request-end"] = time.time()
            self.metadata["http-code"] = status_code
            if self.forced or self.oops.wanted:
                if exc_info and exc_info[0]:
                    root_logger = logging.getLogger()
                    forced = self.forced and 'Forced' or ''
                    message = "Dump of OOPS %s %s" % (self.oops.id, forced)
                    root_logger.error(message, exc_info=exc_info)
                self.serial.dump(self.oops)
        return self.response(status, response_headers, exc_info)

